/* $Id: cpl_imagelist_basic-test.c,v 1.41 2009/02/03 14:18:11 lbilbao Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: lbilbao $
 * $Date: 2009/02/03 14:18:11 $
 * $Revision: 1.41 $
 * $Name: cpl-5_0_0 $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <float.h>

#include "cpl_imagelist_basic.h"
#include "cpl_imagelist_io.h"
#include "cpl_image_gen.h"
#include "cpl_image_bpm.h"
#include "cpl_image_io.h"
#include "cpl_stats.h" 
#include "cpl_msg.h"
#include "cpl_test.h"
#include "cpl_memory.h"
#include "cpl_math_const.h"

#include "cpl_polynomial.h"

/*-----------------------------------------------------------------------------
                                  Define
 -----------------------------------------------------------------------------*/

#define     IMAGESZ         10
#define     NFRAMES         10
#define     MAGIC_NUMBER    42

#define     IMAGESZBIG      512

#define     TEST            10.0
#define     BASE            10.0
#define     POW              2.0

/*-----------------------------------------------------------------------------
                            Private Function prototypes
 -----------------------------------------------------------------------------*/

static void cpl_imagelist_collapse_median_create_tests(int);
static void cpl_imagelist_collapse_median_create_bench(int);

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    cpl_imagelist   *   imlist1 ;
    cpl_imagelist   *   imlist2 ;
    cpl_image       *   img ;
    int                 nbad ;
    int                 bpm_ind, bpm_x, bpm_y ;
    int                 i, j ;
    cpl_boolean         do_bench;

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    /* Insert tests below */

    do_bench = cpl_msg_get_level() <= CPL_MSG_INFO ? CPL_TRUE : CPL_FALSE;

    /* Test cpl_imagelist_swap_axis_create() */
    img = cpl_image_fill_test_create(IMAGESZ, 2*IMAGESZ) ;
    imlist1 = cpl_imagelist_new() ;
    for (i=0 ; i<3*IMAGESZ ; i++) {
        cpl_test_zero(cpl_imagelist_set(imlist1,cpl_image_duplicate(img),i));
    }
    cpl_image_delete(img) ;

    imlist2 = cpl_imagelist_swap_axis_create(imlist1, CPL_SWAP_AXIS_XZ);
    cpl_test_nonnull(imlist2);
    cpl_imagelist_delete(imlist2);

    imlist2 = cpl_imagelist_swap_axis_create(imlist1, CPL_SWAP_AXIS_YZ);
    cpl_test_nonnull(imlist2);
    cpl_imagelist_delete(imlist2);

    /* Test cpl_imagelist_swap_axis_create() with bad pixels */
    cpl_image_reject(cpl_imagelist_get(imlist1, 0), 1, 1) ;
    cpl_image_reject(cpl_imagelist_get(imlist1, 1), 2, 2) ;
    cpl_image_reject(cpl_imagelist_get(imlist1, 2), 2, 5) ;
    imlist2 = cpl_imagelist_swap_axis_create(imlist1, CPL_SWAP_AXIS_XZ);
    cpl_test_nonnull(imlist2);

    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2, 0)) == 1);
    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2, 1)) == 2);
    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2, 2)) == 0);
    cpl_imagelist_delete(imlist2);

    imlist2 = cpl_imagelist_swap_axis_create(imlist1, CPL_SWAP_AXIS_YZ);
    cpl_test_nonnull(imlist2);
    cpl_imagelist_delete(imlist1);

    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2,0)) == 1);
    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2,1)) == 1);
    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2,2)) == 0);
    cpl_test(cpl_image_count_rejected(cpl_imagelist_get(imlist2,4)) == 1);
    cpl_imagelist_delete(imlist2);

    /* Create two images set */
    img = cpl_image_fill_test_create(IMAGESZ, IMAGESZ) ;
    imlist1 = cpl_imagelist_new() ;
    imlist2 = cpl_imagelist_new() ;
    for (i=0 ; i<NFRAMES ; i++) {
        cpl_test_zero(cpl_imagelist_set(imlist1,cpl_image_duplicate(img),i));
        cpl_test_zero(cpl_imagelist_set(imlist2,cpl_image_duplicate(img),i));
    }
    cpl_image_delete(img) ;
    
    /* Test basic operations between images sets */
    cpl_msg_info("","Compute ((((imlist1+imlist2)*imlist2)-imlist2)/imlist2)");
    cpl_test( cpl_imagelist_add(imlist1, imlist2) == CPL_ERROR_NONE );
    cpl_test( cpl_imagelist_multiply(imlist1, imlist2)==CPL_ERROR_NONE );
    cpl_test( cpl_imagelist_subtract(imlist1, imlist2)==CPL_ERROR_NONE );
    cpl_test( cpl_imagelist_divide(imlist1, imlist2) == CPL_ERROR_NONE);
    cpl_imagelist_delete(imlist2) ;

    /* Test basic operations between an image set and an image */
    cpl_msg_info("","Compute ((((imlist1+img)*img)-img)/img)");
    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_DOUBLE) ;
    cpl_test( cpl_image_fill_noise_uniform(img, -1, 1)==CPL_ERROR_NONE );
    cpl_test( cpl_imagelist_add_image(imlist1, img)==CPL_ERROR_NONE );
    cpl_test( cpl_imagelist_multiply_image(imlist1,img)==CPL_ERROR_NONE);
    cpl_test( cpl_imagelist_subtract_image(imlist1,img)==CPL_ERROR_NONE);
    cpl_test( cpl_imagelist_divide_image(imlist1, img)==CPL_ERROR_NONE );
    cpl_image_delete(img) ;

    /* Test basic operations between an image set and a constant */
    cpl_msg_info("","Compute ((((imlist1+42)*42)-42)/42)");
    cpl_test_zero( cpl_imagelist_add_scalar(imlist1, MAGIC_NUMBER) );
    cpl_test_zero( cpl_imagelist_multiply_scalar(imlist1, MAGIC_NUMBER) );
    cpl_test_zero( cpl_imagelist_subtract_scalar(imlist1, MAGIC_NUMBER) );
    cpl_test_zero( cpl_imagelist_divide_scalar(imlist1, MAGIC_NUMBER) );

    /* Test cpl_imagelist_normalise() */
    cpl_msg_info("","Normalize imlist1");
    if (cpl_imagelist_normalise(imlist1, CPL_NORM_MEAN) != CPL_ERROR_NONE) {
        cpl_test_end(1);
    }

    /* Test cpl_imagelist_threshold() */
    cpl_msg_info("","Threshold imlist1");
    cpl_test( cpl_imagelist_threshold(imlist1, 1.0, DBL_MAX, 1.0, 0.0) 
            == CPL_ERROR_NONE );
    
    /* Add bad pixel maps in the images of imlist1 */
    /* image1's bad pixels : 99  */
    /* image2's bad pixels : 49, 99  */
    /* image3's bad pixels : 32, 49, 99  */
    /* image4's bad pixels : 24, 32, 49, 99  */
    /* image5's bad pixels : 19, 24, 32, 49, 99  */
    /* image6's bad pixels : 15, 19, 24, 32, 49, 99  */
    /* image7's bad pixels : 13, 15, 19, 24, 32, 49, 99  */
    /* image8's bad pixels : 11, 13, 15, 19, 24, 32, 49, 99  */
    /* image9's bad pixels : 10, 11, 13, 15, 19, 24, 32, 49, 99  */
    /* image10's bad pixels : 9, 10, 11, 13, 15, 19, 24, 32, 49, 99  */
    for (i=0 ; i < cpl_imagelist_get_size(imlist1) ; i++) {
        const int nx =  cpl_image_get_size_x(cpl_imagelist_get(imlist1, i));
        const int ny =  cpl_image_get_size_y(cpl_imagelist_get(imlist1, i));
        nbad = i + 1 ;
        cpl_test( nbad <= nx*ny );

        for (j=0 ; j<nbad ; j++) {
            bpm_ind = (int)((nx*ny)/(j+1)) - 1 ;
            bpm_x = (bpm_ind % nx) + 1 ;
            bpm_y = (int)(bpm_ind/nx) + 1 ;
            cpl_test_zero( cpl_image_reject(cpl_imagelist_get(imlist1, i),
                                                 bpm_x, bpm_y) );
        }
    }

    /* Test cpl_image_new_from_accepted() */
    cpl_msg_info("","Get the contribution map of imlist1") ;
    cpl_test( img=cpl_image_new_from_accepted(imlist1) );
    cpl_image_delete(img) ;

    /* Test cpl_imagelist_collapse_create() */
    cpl_msg_info("","Average imlist1 to one image") ;
    cpl_test( img=cpl_imagelist_collapse_create(imlist1) );
    cpl_test( cpl_image_count_rejected(img) == 1 );
    cpl_image_delete(img) ;

    /* Test cpl_imagelist_collapse_minmax_create() */
    cpl_msg_info("","Average with rejection on imlist1") ;
    cpl_test( img=cpl_imagelist_collapse_minmax_create(imlist1, 0, 0));
    cpl_image_delete(img) ;

    /* Test cpl_imagelist_collapse_sigclip_create() */
    cpl_msg_info("","Average with sigma clipping rejection on imlist1") ;
    cpl_test(
            img=cpl_imagelist_collapse_sigclip_create(imlist1,
	            0.1, 0.0, 1.0, 1, NULL));
    cpl_image_delete(img) ;
    cpl_imagelist_delete(imlist1) ;

    /* Test cpl_imagelist_collapse_median_create() */
    cpl_imagelist_collapse_median_create_tests(NFRAMES);
    /* Test cpl_imagelist_collapse_median_create() */
    cpl_imagelist_collapse_median_create_tests(NFRAMES*NFRAMES);

    if (do_bench) {
        /* cpl_msg_set_component_on(); */
        cpl_imagelist_collapse_median_create_bench(5);
        cpl_imagelist_collapse_median_create_bench(10);
        cpl_imagelist_collapse_median_create_bench(20);
        cpl_imagelist_collapse_median_create_bench(50);
        cpl_imagelist_collapse_median_create_bench(100);
        /* cpl_msg_set_component_off(); */
    }

    /* Test cpl_imagelist_{exponential, power, logarithm}() */
    {
	float pow_result = powf((float)TEST, (float)POW);
        float log_result = log10f((float)TEST);

	/* Power */
	imlist1 = cpl_imagelist_new();
	imlist2 = cpl_imagelist_new();

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, (float)TEST));
	    cpl_test( !cpl_imagelist_set(imlist1, img, i));
	}

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, pow_result));
	    cpl_test( !cpl_imagelist_set(imlist2, img, i));
	}

	cpl_test( !cpl_imagelist_power(imlist1, (float)POW));

	for (i = 0; i < NFRAMES; i++) {
	    cpl_test_image_abs(cpl_imagelist_get(imlist1, i),
			       cpl_imagelist_get(imlist2, i), 0.0);
	}
	cpl_imagelist_delete(imlist1);
	cpl_imagelist_delete(imlist2);

	/* Logarithm */
	imlist1 = cpl_imagelist_new();
	imlist2 = cpl_imagelist_new();

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, (float)TEST));
	    cpl_test( !cpl_imagelist_set(imlist1, img, i));
	}

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, log_result));
	    cpl_test( !cpl_imagelist_set(imlist2, img, i));
	}

	cpl_test( !cpl_imagelist_logarithm(imlist1, (float)BASE));

	for (i = 0; i < NFRAMES; i++) {
	    cpl_test_image_abs(cpl_imagelist_get(imlist1, i),
			       cpl_imagelist_get(imlist2, i), 0.0);
	}
	cpl_imagelist_delete(imlist1);
	cpl_imagelist_delete(imlist2);

	/* Exponential */
	imlist1 = cpl_imagelist_new();
	imlist2 = cpl_imagelist_new();

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, (float)POW));
	    cpl_test( !cpl_imagelist_set(imlist1, img, i));
	}

	for (i = 0; i < NFRAMES; i++) {
	    img = cpl_image_new(IMAGESZ, IMAGESZ, CPL_TYPE_FLOAT);
	    cpl_test( !cpl_image_add_scalar(img, (float)CPL_MATH_E));
	    cpl_test( !cpl_imagelist_set(imlist2, img, i));
	}

	cpl_test( !cpl_imagelist_exponential(imlist1, (float)CPL_MATH_E));
	cpl_test( !cpl_imagelist_power(imlist2, (float)POW));

	for (i = 0; i < NFRAMES; i++) {
	    cpl_test_image_abs(cpl_imagelist_get(imlist1, i),
			       cpl_imagelist_get(imlist2, i), 0.0);
	}
	cpl_imagelist_delete(imlist1);
	cpl_imagelist_delete(imlist2);
    }

    /* End of tests */
    return cpl_test_end(0);
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Unit test the CPL function
  @param    mz  Number of planes to test
  @return   void

 */
/*----------------------------------------------------------------------------*/
static void cpl_imagelist_collapse_median_create_tests(int mz)
{

    cpl_type  type = CPL_TYPE_DOUBLE;
    const int nx = IMAGESZ;
    const int ny = IMAGESZBIG;
    const int nz = mz | 1; /* Test only odd sizes */
    int       itype;


    cpl_msg_info(cpl_func, "Testing with %d %d X %d images", nz, nx, ny);

    for (itype = 0; itype < 2; itype++, type = CPL_TYPE_FLOAT) {
        cpl_imagelist * imglist = cpl_imagelist_new();
        cpl_image     * image;
        int             i;

        /* Fill imagelist with pixel-type type */
        /* The median of every pixel is zero */
        for (i = 0; i < nz; i++) {
            image = cpl_image_new(nx, ny, type);
            if (i > nz/2) {
                cpl_test_zero(cpl_image_fill_noise_uniform(image, 1.0, (double)nz));
            } else if (i > 0) {
                cpl_test_zero(cpl_image_fill_noise_uniform(image, (double)-nz,
                                                       -1.0));
            }
            cpl_test_zero(cpl_imagelist_set(imglist, image, i));
        }

        cpl_test(cpl_imagelist_get_size(imglist) == nz);

        image = cpl_imagelist_collapse_median_create(imglist);

        cpl_test_nonnull( image);
        cpl_test_error( CPL_ERROR_NONE);

        cpl_test_leq(cpl_image_get_min(image), 0.0);
        cpl_test_leq(cpl_image_get_max(image), 0.0);

        cpl_image_delete(image);
        cpl_imagelist_delete(imglist);

    }

    return;

}


/*----------------------------------------------------------------------------*/
/**
  @brief    Benchmark the CPL function
  @param    mz  Number of planes to test
  @return   void
 */
/*----------------------------------------------------------------------------*/
static void cpl_imagelist_collapse_median_create_bench(int mz)
{

    cpl_imagelist * imglist = cpl_imagelist_new();
    cpl_image     * image;
    const int       nz = mz;
    const int       nx = 2*IMAGESZBIG;
    const int       ny = 2*IMAGESZBIG;
    double          secs;
    int i;


    cpl_msg_info(cpl_func, "Benchmarking with %d %d X %d images", nz, nx, ny);

    /* Fill imagelist */
    for (i = 0; i < nz; i++) {
        image = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
        cpl_test_zero(cpl_image_fill_noise_uniform(image, 0.0, 1.0));
        cpl_test_zero(cpl_imagelist_set(imglist, image, i));
    }

    cpl_test(cpl_imagelist_get_size(imglist) == nz);

    cpl_msg_info(cpl_func, "Testing with %d %d X %d images", nz, nx, ny);

    secs  = cpl_test_get_cputime();

    image = cpl_imagelist_collapse_median_create(imglist);

    secs = cpl_test_get_cputime() - secs;

    cpl_test_nonnull( image);
    cpl_test_error( CPL_ERROR_NONE);
    cpl_image_delete(image);

    cpl_msg_info(cpl_func, "Time spent collapsing with %d X %d X %d: "
                 "[s]: %g", nz, nx, ny, secs);

    cpl_imagelist_delete(imglist);

    return;     

}
